#include <stdio.h>
#include <errno.h>
#include <ldap.h>
#include <pthread.h>

#define BASEDN	"o=Netscape Commuications Corp., c=US"
#define MODDN	"cn=Tim Howes, o=Netscape Commuications Corp., c=US"

static void thread_specific_setup( void );
void *search_thread( void * );
void *mod_thread( void * );
void *my_mutex_alloc( void );
void my_mutex_free( void *mutex );
int my_mutex_lock( void *mutex );
int my_mutex_unlock( void *mutex );
void my_set_errno( int err );
int my_get_errno( void );
void my_set_lderrno( int err, char *matched, char *errmsg, void *dummy );
int my_get_lderrno( char **matched, char **errmsg, void *dummy );

struct ldap_error {
	int	le_errno;
	char	*le_matched;
	char	*le_errmsg;
};
pthread_key_t	key;	/* for per-thread data */
LDAP		*ld;	/* shared LDAP session handle */

main( int argc, char **argv )
{
	struct ldap_thread_fns	tfns;
	pthread_t		search_tid, mod_tid;
	void			*status;

	if ( (ld = ldap_init( "ggood", LDAP_PORT ))
	   == NULL ) {
		perror( "ldap_init" );
		exit( 1 );
	}
	if ( pthread_key_create( &key, NULL ) != 0 ) {
		perror( "pthread_key_create" );
		exit( 1 );
	}
	thread_specific_setup();
	tfns.ltf_mutex_alloc = my_mutex_alloc;
	tfns.ltf_mutex_free = my_mutex_free;
	tfns.ltf_mutex_lock = my_mutex_lock;
	tfns.ltf_mutex_unlock = my_mutex_unlock;
	tfns.ltf_set_errno = my_set_errno;
	tfns.ltf_get_errno = my_get_errno;
	tfns.ltf_set_lderrno = my_set_lderrno;
	tfns.ltf_get_lderrno = my_get_lderrno;
	if ( ldap_set_option( ld, LDAP_OPT_THREAD_FN_PTRS, &tfns )
	   != 0 ) {
		ldap_perror( ld, "ldap_set_option" );
		exit( 1 );
	}
	if ( ldap_simple_bind_s( ld, NULL, NULL ) != LDAP_SUCCESS ) {
		ldap_perror( ld, "ldap_simple_bind_s" );
		exit( 1 );
	}

	if ( pthread_create( &search_tid, NULL, search_thread, NULL )
	   != 0 || pthread_create( &mod_tid, NULL, mod_thread, NULL )
	   != 0 ) {
		perror( "pthread_create" );
		exit( 1 );
	}
	pthread_join( search_tid, &status );
	pthread_join( mod_tid, &status );
	return( 0 );
}

static void
thread_specific_setup( void )
{
	void	*tsd;

	tsd = (void *) calloc( 1, sizeof(struct ldap_error) );
	pthread_setspecific( key, tsd );
}

void *
search_thread( void *arg )
{
	int		i;
	LDAPMessage	*result;

	thread_specific_setup();
	for ( i = 0; i < 100; i++ ) {
		if ( ldap_search_s( ld, BASEDN, LDAP_SCOPE_SUBTREE,
		   "(sn=Jensen)", NULL, 0, &result ) != LDAP_SUCCESS ) {
			ldap_perror( ld, "ldap_search_s" );
			exit( 1 );
		}
		/* ...do something with the results... */
		ldap_msgfree( result );
	}
}

void *
mod_thread( void *arg )
{
	int	i;
 	LDAPMod	**mods;

	thread_specific_setup();

	/*  fill in mods  */
 	for ( i = 0; i < 100; i++ ) {
		if ( ldap_modify_s( ld, MODDN, mods ) != LDAP_SUCCESS ) {
			ldap_perror( ld, "ldap_modify_s" );
			exit( 1 );
		}
	}
}
void *
my_mutex_alloc( void )
{
	pthread_mutex_t	*mutex;

	mutex = (pthread_mutex_t *) malloc( sizeof(pthread_mutex_t) );
	pthread_mutex_init( mutex, NULL );
	return( (void *) mutex );
}

void
my_mutex_free( void *mutex )
{
	pthread_mutex_destroy( (pthread_mutex_t *) mutex );
 	free( mutex );
}

int
my_mutex_lock( void *mutex )
{
	return( pthread_mutex_lock( (pthread_mutex_t *)mutex ) );
}

int
my_mutex_unlock( void *mutex )
{
	return( pthread_mutex_unlock( (pthread_mutex_t *)mutex ) );
}

void
my_set_errno( int err )
{
	errno = err;
}

int
my_get_errno( void )
{
	return( errno );
}

void
my_set_lderrno( int err, char *matched, char *errmsg, void *dummy )
{
	struct ldap_error	*le;

	le = pthread_getspecific( key );
	le->le_errno = err;
	if ( le->le_matched != NULL ) {
		ldap_memfree( le->le_matched );
	}
	le->le_matched = matched;
	if ( le->le_errmsg != NULL ) {
		ldap_memfree( le->le_errmsg );
	}
	le->le_errmsg = errmsg;
}

int
my_get_lderrno( char **matched, char **errmsg, void *dummy )
{
	struct ldap_error	*le;

	le = pthread_getspecific( key );
	if ( matched != NULL ) {
		*matched = le->le_matched;
	}
	if ( errmsg != NULL ) {
		*errmsg = le->le_errmsg;
	}
	return( le->le_errno );
}
